from typing import Generic, TypeVar

from rainbowadn.chain.abstractreductionchainmetafactory import AbstractReductionChainMetaFactory
from rainbowadn.chain.block import Block
from rainbowadn.chain.blockchain import BlockChainFactory
from rainbowadn.chain.chaincollectionfactory import ChainCollectionFactory
from rainbowadn.chain.reduction.reduction import Reduction
from rainbowadn.chain.reduction.reductionchainprotocol import ReductionChainProtocol
from rainbowadn.chain.reduction.reductionprotocol import ReductionProtocol
from rainbowadn.chain.stages.stage import StateStage
from rainbowadn.core.rainbow_factory import RainbowFactory

__all__ = ('ReductionChainMetaFactory',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class ReductionChainMetaFactory(
    AbstractReductionChainMetaFactory[
        Block[
            ReductorType,
            StateStage[ReductorType, AccumulatorType, Reduction[ReductorType, AccumulatorType]]
        ],
        ReductorType,
        AccumulatorType
    ],
    Generic[ReductorType, AccumulatorType]
):
    def factory(
            self,
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
            protocol: ReductionProtocol[ReductorType, AccumulatorType],
    ) -> ChainCollectionFactory[
        Block[
            ReductorType,
            StateStage[ReductorType, AccumulatorType, Reduction[ReductorType, AccumulatorType]]
        ],
        ReductorType,
        AccumulatorType
    ]:
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        assert isinstance(protocol, ReductionProtocol)
        return BlockChainFactory(
            ReductionChainProtocol(
                protocol,
                reductor_factory,
                accumulator_factory,
            )
        )

    def loose(self) -> AbstractReductionChainMetaFactory[
        Block[
            ReductorType,
            StateStage[ReductorType, AccumulatorType, Reduction[ReductorType, AccumulatorType]]
        ],
        ReductorType,
        AccumulatorType
    ]:
        return self
