from typing import Generic, TypeVar

from rainbowadn.chain.reduction.reduced import Reduced
from rainbowadn.chain.reduction.reduction import Reduction
from rainbowadn.chain.reduction.reductionprotocol import ReductionProtocol
from rainbowadn.chain.reduction.reductionresult import ReductionResult
from rainbowadn.chain.stages.derivation.activestageprotocol import ActiveStageProtocol
from rainbowadn.chain.stages.derivation.derived import Derived
from rainbowadn.chain.stages.derivation.derivedstage import DerivedStage
from rainbowadn.chain.stages.derivation.derivedstate import DerivedState
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.nullability.nullablereference import NullableReference

__all__ = ('ReductionStageProtocol',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class ReductionStageProtocol(
    ActiveStageProtocol[ReductorType, AccumulatorType, Reduction[ReductorType, AccumulatorType]],
    Generic[ReductorType, AccumulatorType]
):
    def __init__(self, protocol: ReductionProtocol[ReductorType, AccumulatorType]):
        assert isinstance(protocol, ReductionProtocol)
        self.protocol = protocol
        super().__init__()

    async def _derive_accumulator(
            self,
            previous: NullableReference[AccumulatorType],
    ) -> HashPoint[AccumulatorType]:
        assert isinstance(previous, NullableReference)
        if previous.null():
            return self.protocol.initial(previous.factory)
        else:
            return await self.protocol.header_filter(previous.hashpoint())

    async def derive_header(
            self,
            previous: NullableReference[AccumulatorType],
            header: HashPoint[ReductorType]
    ) -> HashPoint[Reduction[ReductorType, AccumulatorType]]:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        return HashPoint.of(
            Reduction(
                header,
                await self._derive_accumulator(previous)
            )
        )

    @classmethod
    def _derived_from_reduced(
            cls,
            reduced: ReductionResult[ReductorType, AccumulatorType]
    ) -> Derived[ReductorType, Reduction[ReductorType, AccumulatorType]]:
        assert isinstance(reduced, ReductionResult)
        if isinstance(reduced, Reduction):
            return DerivedStage(HashPoint.of(reduced))
        elif isinstance(reduced, Reduced):
            return DerivedState(reduced.reduced)
        else:
            raise TypeError

    async def derive_stage_or_state(
            self,
            stage: HashPoint[Reduction[ReductorType, AccumulatorType]]
    ) -> Derived[ReductorType, Reduction[ReductorType, AccumulatorType]]:
        assert isinstance(stage, HashPoint)
        return self._derived_from_reduced(await self.protocol.reduce(await stage.resolve()))
