from typing import TypeVar

from rainbowadn.chain.stages.derivation.activestageprotocol import ActiveStageProtocol
from rainbowadn.chain.stages.stage import StateStage
from rainbowadn.chain.states.activestateprotocol import ActiveStateProtocol
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.nullability.nullablereference import NullableReference
from rainbowadn.core.rainbow_factory import RainbowFactory

__all__ = ('ActiveStageStateProtocol',)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class ActiveStageStateProtocol(
    ActiveStateProtocol[
        HeaderType,
        StateStage[
            HeaderType,
            BaseStateType,
            StageType
        ]
    ]
):
    def __init__(
            self,
            protocol: ActiveStageProtocol,
            stage_factory: RainbowFactory[StageType],
            base_state_factory: RainbowFactory[BaseStateType],
    ):
        assert isinstance(protocol, ActiveStageProtocol)
        assert isinstance(stage_factory, RainbowFactory)
        assert isinstance(base_state_factory, RainbowFactory)
        self.protocol = protocol
        self.stage_factory = stage_factory
        self.base_state_factory = base_state_factory

    async def derive(
            self,
            previous: NullableReference[
                StateStage[
                    HeaderType,
                    BaseStateType,
                    StageType
                ]
            ],
            header: HashPoint[HeaderType],
    ) -> HashPoint[
        StateStage[
            HeaderType,
            BaseStateType,
            StageType
        ]
    ]:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        return await self.protocol.derive_full(
            previous,
            header,
            self.base_state_factory,
            self.stage_factory
        )
