from typing import Generic, Iterable, TypeVar

from rainbowadn.core.hash_point_format import hash_point_format, tabulate
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.core.recursivementionable import RecursiveMentionable
from rainbowadn.core.resolverorigin import ResolverOrigin

__all__ = ('Array', 'ArrayFactory',)

ElementType = TypeVar('ElementType')


class Array(RecursiveMentionable, Generic[ElementType]):
    def __init__(
            self,
            factory: RainbowFactory[ElementType],
            array: tuple[HashPoint[ElementType], ...]
    ):
        assert isinstance(factory, RainbowFactory)
        assert isinstance(array, tuple)
        self.factory = factory
        self.array = array

    def points(self) -> Iterable[HashPoint]:
        return self.array

    def __bytes__(self):
        return b''.join(map(HashPoint.__bytes__, self.array))

    def __factory__(self) -> RainbowFactory['Array']:
        return ArrayFactory(self.factory)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        formatted = f'('
        for hash_point in self.array:
            formatted += f'{tabulate(tab + 1)}{await hash_point_format(hash_point, tab + 1)}'
        return f'{formatted}' \
               f'{tabulate(tab)})'


class ArrayFactory(RainbowFactory[Array], Generic[ElementType]):
    def __init__(
            self,
            factory: RainbowFactory[ElementType]
    ):
        assert isinstance(factory, RainbowFactory)
        self.factory = factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> Array:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return Array(
            self.factory,
            tuple(
                ResolverOrigin(self.factory, source[i:i + HashPoint.HASH_LENGTH], resolver).hash_point()
                for
                i
                in
                range(0, len(source), HashPoint.HASH_LENGTH)
            )
        )
