from typing import Generic, TypeVar

from rainbowadn.data.collection.trees.binary.binaryprotocolized import BinaryProtocolized

__all__ = ('BinaryAction',)

from rainbowadn.data.collection.trees.binary.protocolizedbinarysplit import ProtocolizedBinarySplit

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')
ActionType = TypeVar('ActionType')


class BinaryAction(Generic[ActiveKeyType, MetaDataType, TreeType, ActionType]):
    async def on(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType],
    ) -> ActionType:
        if (split := await ProtocolizedBinarySplit.split_of(protocolized)) is None:
            return await self.on_null(protocolized)
        else:
            return await self.on_split(split)

    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType],
    ) -> ActionType:
        raise NotImplementedError

    async def on_split(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> ActionType:
        raise NotImplementedError
