import abc
from typing import Generic, TypeVar

from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.data.collection.trees.binary.actions.binaryaction import BinaryAction
from rainbowadn.data.collection.trees.binary.protocolizedbinarysplit import ProtocolizedBinarySplit
from rainbowadn.data.collection.trees.comparison.comparator import Comparison, Equal, Left, Right

__all__ = ('CompareAction',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')
ActionType = TypeVar('ActionType')


class CompareAction(
    BinaryAction[ActiveKeyType, MetaDataType, TreeType, ActionType],
    Generic[ActiveKeyType, MetaDataType, TreeType, ActionType],
    abc.ABC
):
    def __init__(self, key: HashPoint[ActiveKeyType]):
        assert isinstance(key, HashPoint)
        self.key = key

    async def on_split(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> ActionType:
        assert isinstance(case, ProtocolizedBinarySplit)
        comparison: Comparison = await case.protocol.comparator.compare(case.split.key, self.key)
        assert isinstance(comparison, Comparison)
        if isinstance(comparison, Equal):
            return await self.on_equal(case, comparison)
        elif isinstance(comparison, Left):
            return await self.on_left(case)
        elif isinstance(comparison, Right):
            return await self.on_right(case)
        else:
            raise TypeError

    async def on_equal(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType],
            equal: Equal
    ) -> ActionType:
        raise NotImplementedError

    async def on_left(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> ActionType:
        raise NotImplementedError

    async def on_right(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> ActionType:
        raise NotImplementedError
