from typing import Generic, TypeVar

from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.data.collection.trees.binary.binarycreation import BinaryCreation
from rainbowadn.data.collection.trees.binary.binaryprotocolized import BinaryProtocolized
from rainbowadn.data.collection.trees.binary.binarysplit import BinarySplit

__all__ = ('Symmetric', 'InnerOuter', 'OuterInner',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class Symmetric(
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(
            self,
            protocol: BinaryCreation[ActiveKeyType, MetaDataType, TreeType],
    ):
        self.protocol = protocol

    def inner(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        raise NotImplementedError

    def outer(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        raise NotImplementedError

    async def tree(
            self,
            inner: TreeType,
            outer: TreeType,
            key: HashPoint[ActiveKeyType]
    ) -> TreeType:
        raise NotImplementedError

    def protocolizedi(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType]:
        return BinaryProtocolized(self.protocol, self.inner(split))

    def protocolizedo(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType]:
        return BinaryProtocolized(self.protocol, self.outer(split))


class InnerOuter(Symmetric):
    def inner(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return split.treel

    def outer(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return split.treer

    async def tree(
            self,
            inner: TreeType,
            outer: TreeType,
            key: HashPoint[ActiveKeyType]
    ) -> TreeType:
        return await self.protocol.tree(inner, outer, key)


class OuterInner(Symmetric):
    def inner(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return split.treer

    def outer(
            self,
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        return split.treel

    async def tree(
            self,
            inner: TreeType,
            outer: TreeType,
            key: HashPoint[ActiveKeyType]
    ) -> TreeType:
        return await self.protocol.tree(outer, inner, key)
