import abc
from typing import Generic, TypeVar

from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.data.collection.trees.binary.binarybalancing import BinaryBalancing
from rainbowadn.data.collection.trees.binary.binarycreation import BinaryCreation

__all__ = ('BalancedCreation',)

from rainbowadn.data.collection.trees.binary.binaryprotocolized import BinaryProtocolized

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class BalancedCreation(
    BinaryCreation[ActiveKeyType, MetaDataType, TreeType],
    Generic[ActiveKeyType, MetaDataType, TreeType],
    abc.ABC
):
    def __init__(
            self,
            protocol: BinaryBalancing[ActiveKeyType, MetaDataType, TreeType]
    ):
        assert isinstance(protocol, BinaryBalancing)
        self.protocol = protocol
        super().__init__(protocol.comparator)

    async def _tree(self, treel: TreeType, treer: TreeType, key: HashPoint[ActiveKeyType]) -> TreeType:
        raise NotImplementedError

    async def tree(self, treel: TreeType, treer: TreeType, key: HashPoint[ActiveKeyType]) -> TreeType:
        assert isinstance(key, HashPoint)
        return await self.protocol.balance(
            BinaryProtocolized(
                self,
                await self._tree(
                    await self.protocol.balance(BinaryProtocolized(self, treel)),
                    await self.protocol.balance(BinaryProtocolized(self, treer)),
                    key
                )
            )
        )
