from typing import Generic, Optional, TypeVar

from rainbowadn.data.collection.trees.binary.binarycreation import BinaryCreation
from rainbowadn.data.collection.trees.binary.binaryprotocolized import BinaryProtocolized
from rainbowadn.data.collection.trees.binary.binarysplit import BinarySplit

__all__ = ('ProtocolizedBinarySplit',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class ProtocolizedBinarySplit(
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(
            self,
            protocol: BinaryCreation[ActiveKeyType, MetaDataType, TreeType],
            split: BinarySplit[ActiveKeyType, MetaDataType, TreeType],
            tree: TreeType
    ):
        assert isinstance(protocol, BinaryCreation)
        assert isinstance(split, BinarySplit)
        self.protocol = protocol
        self.split = split
        self.tree = tree

    def protocolized(self) -> BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType]:
        return BinaryProtocolized(self.protocol, self.tree)

    def protocolizedl(self) -> BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType]:
        return BinaryProtocolized(self.protocol, self.split.treel)

    def protocolizedr(self) -> BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType]:
        return BinaryProtocolized(self.protocol, self.split.treer)

    @classmethod
    async def split_of(
            cls,
            protocolized: BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType]
    ) -> Optional[
        'ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]'
    ]:
        if (split := await protocolized.split()) is None:
            return None
        else:
            return ProtocolizedBinarySplit(protocolized.protocol, split, protocolized.tree)
