from typing import Generic, TypeVar

from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.data.collection.keyed import Keyed
from rainbowadn.data.collection.trees.comparison.comparator import Comparator, Comparison

__all__ = ('KeyedComparator',)

ComparatorKeyType = TypeVar('ComparatorKeyType')


class KeyedComparator(
    Comparator[Keyed[ComparatorKeyType]], Generic[ComparatorKeyType]
):
    def __init__(self, comparator: Comparator[ComparatorKeyType]):
        assert isinstance(comparator, Comparator)
        self.comparator = comparator
        super().__init__()

    async def compare(
            self,
            original: HashPoint[Keyed[ComparatorKeyType]],
            key: HashPoint[Keyed[ComparatorKeyType]]
    ) -> Comparison:
        assert isinstance(original, HashPoint)
        assert isinstance(key, HashPoint)
        return await self.comparator.compare(
            (await original.resolve()).key,
            (await key.resolve()).key,
        )
