from collections import OrderedDict
from typing import MutableMapping

from rainbowadn.core.extendableresolver import ExtendableResolver, Mentioned
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.mentionable import Mentionable
from rainbowadn.core.recursivementionable import RecursiveMentionable


class DictResolver(ExtendableResolver):
    def __init__(self):
        self.table: MutableMapping[bytes, bytes] = OrderedDict()

    async def resolve(self, point: bytes) -> tuple[bytes, 'HashResolver']:
        assert isinstance(point, bytes)
        return self.table[point], self

    async def save(self, hash_point: HashPoint) -> None:
        assert isinstance(hash_point, HashPoint)
        if hash_point.point in self.table:
            pass
        else:
            value: Mentionable = await hash_point.resolve()
            assert isinstance(value, Mentionable)
            self.table[hash_point.point] = HashPoint.bytes_of_mentioned(value)
            if isinstance(value, RecursiveMentionable):
                for hash_point in value.points():
                    await self.save(hash_point)

    async def extend(self, hash_point: HashPoint[Mentioned]) -> 'ExtendableResolver':
        await self.save(hash_point)
        return self
