import functools


class Instrumentation:
    def __init__(self, target, methodname: str):
        self.target = target
        self.methodname = methodname

    def instrument(self, *args, **kwargs):
        raise NotImplementedError

    def __enter__(self):
        assert not hasattr(self, 'method')
        self.method = getattr(self.target, self.methodname)

        @functools.wraps(self.method)
        def wrap(*args, **kwargs):
            self.instrument(*args, **kwargs)
            return self.method(*args, **kwargs)

        self.wrap = wrap

        setattr(self.target, self.methodname, self.wrap)

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert getattr(self.target, self.methodname) is self.wrap
        del self.wrap
        setattr(self.target, self.methodname, self.method)
        del self.method


class Counter(Instrumentation):
    def __init__(self, target, methodname: str):
        super().__init__(target, methodname)
        self.counter = 0

    def instrument(self, *args, **kwargs):
        self.counter += 1
