from typing import Generic, Iterable, TypeVar

from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.hashresolver import HashResolver
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.core.recursivementionable import RecursiveMentionable
from rainbowadn.core.resolverorigin import ResolverOrigin
from rainbowadn.toplevel.thresholdprotocol import ThresholdProtocol

__all__ = ('ValidReference', 'ValidReferenceFactory')

Referenced = TypeVar('Referenced')


class ValidReference(RecursiveMentionable, Generic[Referenced]):
    def __init__(self, reference: HashPoint[Referenced], protocol: ThresholdProtocol[Referenced]):
        assert isinstance(reference, HashPoint)
        assert isinstance(protocol, ThresholdProtocol)
        self.reference = reference
        self.protocol = protocol

    def points(self) -> Iterable[HashPoint]:
        return [self.reference]

    async def resolve(self) -> Referenced:
        referenced: Referenced = await self.reference.resolve()
        assert self.reference.point < (await self.protocol.threshold(referenced))
        return referenced

    def __bytes__(self):
        return bytes(self.reference)

    def __factory__(self) -> RainbowFactory['ValidReference[Referenced]']:
        return ValidReferenceFactory(self.reference.factory, self.protocol)


class ValidReferenceFactory(RainbowFactory[ValidReference[Referenced]], Generic[Referenced]):
    def __init__(self, factory: RainbowFactory[Referenced], protocol: ThresholdProtocol[Referenced]):
        assert isinstance(factory, RainbowFactory)
        assert isinstance(protocol, ThresholdProtocol)
        self.factory = factory
        self.protocol = protocol

    def from_bytes(self, source: bytes, resolver: HashResolver) -> ValidReference[Referenced]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return ValidReference(ResolverOrigin(self.factory, source, resolver).hash_point(), self.protocol)
