from typing import Generic, TypeVar

from rainbowadn.chain.blockchainprotocol import BlockChainProtocol
from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.core.rainbow_factory import RainbowFactory
from rainbowadn.wrisbt.wrisbtindex import WrisbtIndex, WrisbtIndexFactory
from rainbowadn.wrisbt.wrisbtparametres import WrisbtParametres
from rainbowadn.wrisbt.wrisbtprotocol import WrisbtProtocol
from rainbowadn.wrisbt.wrisbtroot import WrisbtRoot, WrisbtRootFactory

__all__ = ('WrisbtChainProtocol',)

TargetType = TypeVar('TargetType')


class WrisbtChainProtocol(
    BlockChainProtocol[
        TargetType,
        WrisbtIndex,
        WrisbtRoot
    ],
    Generic[TargetType]
):
    def __init__(
            self,
            total_factory: RainbowFactory[TargetType],
            keymin: int
    ):
        assert isinstance(total_factory, RainbowFactory)
        assert isinstance(keymin, int)
        assert keymin > 0
        self.keymin = keymin
        super().__init__(
            WrisbtProtocol(keymin),
            total_factory,
            WrisbtIndexFactory(keymin)
        )
        self.total_factory = total_factory

    def actual_state_factory(self) -> RainbowFactory[WrisbtRoot]:
        return WrisbtRootFactory(WrisbtParametres(self.keymin, HashPoint.HASH_LENGTH))

    async def actual_state(self, state: WrisbtIndex) -> HashPoint[WrisbtRoot]:
        assert isinstance(state, WrisbtIndex)
        return state.total
